/**
 * Copyright (c) 2014-2016 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see http://www.gnu.org/licenses/
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.microkernel.registry.overlay;

import java.util.Comparator;
import java.util.Set;
import java.util.TreeSet;

import org.ow2.petals.clientserverapi.configuration.ContainerConfiguration;
import org.ow2.petals.microkernel.api.configuration.ConfigurationService;
import org.ow2.petals.microkernel.api.configuration.exception.ConfigurationException;
import org.ow2.petals.microkernel.api.system.sharedarea.SharedAreaService;
import org.ow2.petals.microkernel.api.util.ConfigurationServiceUtils;
import org.ow2.petals.topology.generated.Container;
import org.ow2.petals.topology.generated.Topology;

import com.ebmwebsourcing.easycommons.log.LoggingUtil;
import com.hazelcast.core.EntryEvent;
import com.hazelcast.map.listener.EntryAddedListener;
import com.hazelcast.map.listener.EntryRemovedListener;
import com.hazelcast.map.listener.EntryUpdatedListener;

/**
 * Listener on the dynamic topology entry. Used to detect added or removed container from our domain.
 * 
 * The container receives only updates of its topology! The topology changes must be applied in the right order, ie. in
 * the same order than containers are attached or detached.
 * 
 * @author Christophe DENEUX - Linagora
 */
public class DynamicTopologyListener implements EntryAddedListener<String, Topology>,
        EntryUpdatedListener<String, Topology>, EntryRemovedListener<String, Topology> {

    private final static Comparator<Container> CONTAINER_COMPARATOR = new Comparator<Container>() {
        @Override
        public int compare(Container o1, Container o2) {
            return o1.getName().compareTo(o2.getName());
        }
    };

    private final ConfigurationService configurationService;

    /**
     * Domain name, used also as topology name
     */
    private final String domainName;

    private final LoggingUtil log;

    public DynamicTopologyListener(final ConfigurationService configurationService, final String domainName,
            final LoggingUtil log) {
        this.configurationService = configurationService;
        this.domainName = domainName;
        this.log = log;
    }

    @Override
    public void entryAdded(final EntryEvent<String, Topology> event) {
        // NOP: Only changed are managed to detect updated topology
        this.log.call();
    }

    @Override
    public void entryRemoved(final EntryEvent<String, Topology> event) {
        // NOP: Only changed are managed to detect updated topology
        this.log.call();
    }

    /**
     * According to Hazelcast 3.5 documentation
     * (http://docs.hazelcast.org/docs/3.5/manual/html/globaleventconfiguration.html):
     * "For map and multimap, the order is preserved for the operations on the same key of the entry." Moreover,
     * {@link SharedAreaService#setTopology(Topology)} uses a mutex to set only one topology at once in the topology map
     * on which this listener is set. So, we will receive here notification in the right order and it is sufficient to
     * synchronize this method to avoid concurrent notification processing.
     */
    @Override
    public synchronized void entryUpdated(final EntryEvent<String, Topology> event) {
        this.log.start();
        try {
            final Topology oldTopology = event.getOldValue();
            if (oldTopology == null) {
                this.log.error("No previous topology was registered in the registry. Please contact Petals ESB Vendor !");
            } else {
                final Topology newTopology = event.getValue();
                // Note we know by design that the old and new topology must have the same name aka the key name!
                if (this.log.isWarnEnabled() && (!event.getKey().equals(oldTopology.getDomain().getName())
                        || !event.getKey().equals(newTopology.getDomain().getName()))) {
                    this.log.warning("Inconsistency between names in the event: expected: '" + this.domainName
                            + "', key: '" + event.getKey() + "', old: '" + oldTopology.getDomain().getName()
                            + "', new: '" + newTopology.getDomain().getName()
                            + "'. This should never happen, but using the expected domain name if it makes sense.");
                }
                
                // this can happen if we receive events for an old topology maybe... let's simply ignore it!
                if (!oldTopology.getDomain().getName().equals(domainName)) {
                    if (this.log.isDebugEnabled()) {
                        this.log.debug("Received an event for domain '" + oldTopology.getDomain().getName()
                                + "' but was expecting '" + this.domainName
                                + "': ignoring, it must be an outdated event.");
                    }
                    return;
                }

                if (this.log.isDebugEnabled()) {
                    this.log.debug("Topology updated in hazelcat cluster, old topology: " + oldTopology.toString());
                    this.log.debug("Topology updated in hazelcat cluster, new topology: " + newTopology.toString());
                }

                // To be able to restart a container in the right topology, we persist the topology here. It's mainly
                // used detaching a container. When attaching a container the persistence is forced to improved error
                // management, see comments in attachment process.
                try {
                    this.configurationService.persistTopology(newTopology);
                } catch (final ConfigurationException e) {
                    this.log.warning(
                            "An error occurs persisting the topology. A restart of the container can be incorrect (not in the right domain), or fail !!",
                            e);
                }

                final String myContainerName = this.configurationService.getContainerConfiguration().getName();

                final Set<Container> added = new TreeSet<>(CONTAINER_COMPARATOR);
                final Set<Container> removed = new TreeSet<>(CONTAINER_COMPARATOR);

                added.addAll(newTopology.getDomain().getContainer());
                // Caution, we can't use "added.removeAll(oldTopology.getDomain().getContainer())" because
                // the implementation of removeAll from Set does not always use the comparator of the collection!
                for (final Container oldTopologyContainer : oldTopology.getDomain().getContainer()) {
                    added.remove(oldTopologyContainer);
                }

                removed.addAll(oldTopology.getDomain().getContainer());
                // Caution, we can't use "removed.removeAll(newTopology.getDomain().getContainer())" because
                // the implementation of removeAll from Set does not always use the comparator of the collection!
                for (final Container newTopologyContainer : newTopology.getDomain().getContainer()) {
                    removed.remove(newTopologyContainer);
                }

                if (this.log.isDebugEnabled()) {
                    this.log.debug("Added containers: " + added.toString());
                    this.log.debug("Removed containers: " + removed.toString());
                }

                for (final Container container : added) {
                    try {
                        if (!myContainerName.equals(container.getName())) {
                            final ContainerConfiguration containerConfig = new ContainerConfiguration();
                            ConfigurationServiceUtils.initializeContainerConfiguration(containerConfig, container,
                                    this.domainName);
                            this.configurationService.addContainerConfiguration(containerConfig);
                        }
                    } catch (final ConfigurationException e) {
                        this.log.warning("An error happened while adding a container detected as new to the domain", e);
                    }
                }

                for (final Container container : removed) {
                    try {
                        if (!myContainerName.equals(container.getName())) {
                            final ContainerConfiguration containerConfig = new ContainerConfiguration();
                            ConfigurationServiceUtils.initializeContainerConfiguration(containerConfig, container,
                                    this.domainName);
                            this.configurationService.removeContainerConfiguration(containerConfig);
                        }
                    } catch (final ConfigurationException e) {
                        this.log.warning(
                                "An error happened while removing a container detected as removed from the domain", e);
                    }
                }
            }
        } finally {
            this.log.end();
        }

    }
}
