/**
 * Copyright (c) 2013-2016 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see http://www.gnu.org/licenses/
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.microkernel.registry.overlay;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertNotNull;
import static org.junit.Assert.assertNull;

import java.net.URISyntaxException;
import java.util.Collection;
import java.util.Collections;

import javax.xml.namespace.QName;

import org.junit.Test;
import org.ow2.easywsdl.extensions.wsdl4complexwsdl.api.WSDL4ComplexWsdlException;
import org.ow2.petals.jbi.servicedesc.endpoint.Location;
import org.ow2.petals.jbi.servicedesc.endpoint.PetalsServiceEndpoint;
import org.ow2.petals.jbi.servicedesc.endpoint.PetalsServiceEndpoint.EndpointType;
import org.ow2.petals.microkernel.api.system.sharedarea.SharedAreaService;
import org.ow2.petals.microkernel.api.system.sharedarea.exception.SharedAreaException;
import org.w3c.dom.Document;

import com.ebmwebsourcing.easycommons.lang.reflect.ReflectionHelper;
import com.ebmwebsourcing.easycommons.uuid.QualifiedUUIDGenerator;

/**
 * Unit tests of {@link RegistryOverlayImpl#leaveTopology()} and {@link RegistryOverlayImpl#joinTopology(Collection)}
 * 
 * @author Christophe DENEUX - Linagora
 * 
 */
public class RegistryOverlayImpl_LeaveJoinToplogy_Test extends AbstractRegistryOverlayImplTest {

    /**
     * A namespace name
     */
    private final static String A_NAMESPACE = "http://petals.ow2.org/unit-tests";

    /**
     * A service name
     */
    private final static QName SERVICE_NAME_1 = new QName(A_NAMESPACE, "service-1");

    /**
     * A interface name
     */
    private final static QName INTERFACE_NAME_1 = new QName(A_NAMESPACE, "interface-1");

    /**
     * A container name
     */
    private final static String CONTAINER_NAME_1 = "container-1";

    /**
     * Another container name
     */
    private final static String CONTAINER_NAME_2 = "container-2";

    /**
     * A endpoint name
     */
    private final static String ENDPOINT_NAME_1;

    /**
     * Another endpoint name
     */
    private final static String ENDPOINT_NAME_2;

    static {
        final QualifiedUUIDGenerator uuidGenerator = new QualifiedUUIDGenerator("petals");
        ENDPOINT_NAME_1 = uuidGenerator.getNewID();
        ENDPOINT_NAME_2 = uuidGenerator.getNewID();
    }

    /**
     * <p>
     * Check the method {@link RegistryOverlayImpl#removeAllLocalEndpoints()}:
     * <ol>
     * <li>in a dedicated topology,</li>
     * <li>an endpoint of local container is registered,</li>
     * <li>an endpoint of a not local container is registered,</li>
     * </ol>
     * </p>
     * <p>
     * Expected result:
     * <ol>
     * <li>no error occurs</li>
     * <li>the endpoint of the local container is no more registered,</li>
     * <li>endpoints of other container is retrieved using {@link SharedAreaService#getEndpoint(QName, String)}</li>
     * <li>endpoints of other container is retrieved using
     * {@link SharedAreaService#getEndpointsForService(QName, EndpointType)}</li>
     * <li>endpoints of other container is retrieved using
     * {@link SharedAreaService#getEndpointsForInterface(QName, EndpointType)}</li>
     * </ol>
     * </p>
     */
    @Test
    public void noLocalEndpointsToRemove() throws Exception {

        final RegistryOverlayImpl localRegistryOverlayImpl = new RegistryOverlayImpl();
        ReflectionHelper.setPrivateField(
                localRegistryOverlayImpl,
                "configurationService",
                RegistryOverlayImpl_LeaveJoinToplogy_Test.createCfgServiceMock(
                        PETALS_REGISTRY_NODE.getGroupName(), PETALS_REGISTRY_NODE.getPassword(),
                        PETALS_REGISTRY_NODE.getHost(), String.valueOf(PETALS_REGISTRY_NODE.getPort()), this.getClass()
                                .getSimpleName() + "_testNoLocalEndpointsToRemove", CONTAINER_NAME_1));

        final RegistryOverlayImpl remoteRegistryOverlayImpl = new RegistryOverlayImpl();
        ReflectionHelper.setPrivateField(
                remoteRegistryOverlayImpl,
                "configurationService",
                RegistryOverlayImpl_LeaveJoinToplogy_Test.createCfgServiceMock(
                        PETALS_REGISTRY_NODE.getGroupName(), PETALS_REGISTRY_NODE.getPassword(),
                        PETALS_REGISTRY_NODE.getHost(), String.valueOf(PETALS_REGISTRY_NODE.getPort()), this.getClass()
                                .getSimpleName() + "_testNoLocalEndpointsToRemove", CONTAINER_NAME_2));
        try {
            localRegistryOverlayImpl.start();
            remoteRegistryOverlayImpl.start();

            localRegistryOverlayImpl.loadTopology(Collections.<SharedAreaService.SavedEndpoint> emptyList());
            remoteRegistryOverlayImpl.loadTopology(Collections.<SharedAreaService.SavedEndpoint> emptyList());

            final Document document = readEndpointDescription("echo.wsdl");

            final Location localLocation = new Location(CONTAINER_NAME_1, "");
            final Location remoteLocation = new Location(CONTAINER_NAME_2, "");

            // Add the local endpoint
            localRegistryOverlayImpl.registerInternalEndpoint(ENDPOINT_NAME_1, SERVICE_NAME_1,
                    new QName[] { INTERFACE_NAME_1 }, localLocation, document, false);

            // Add the remote endpoint
            remoteRegistryOverlayImpl.registerInternalEndpoint(ENDPOINT_NAME_2, SERVICE_NAME_1,
                    new QName[] { INTERFACE_NAME_1 }, remoteLocation, document, false);

            // Check that endpoints are correctly registered
            assertEndpointsRegistered(localRegistryOverlayImpl);

            // ///////////////////////////////
            // Remove all local endpoints
            // ///////////////////////////////
            final Collection<SharedAreaService.SavedEndpoint> removedEndpoints = localRegistryOverlayImpl
                    .leaveTopology();

            // Try to retrieve the local endpoint
            assertNull(localRegistryOverlayImpl.getEndpoint(SERVICE_NAME_1, ENDPOINT_NAME_1));

            // Check that the remote endpoint is always visible
            final PetalsServiceEndpoint retrievedEndpoint = localRegistryOverlayImpl.getEndpoint(SERVICE_NAME_1,
                    ENDPOINT_NAME_2);
            assertNotNull(retrievedEndpoint);
            assertEquals(retrievedEndpoint.getServiceName(), SERVICE_NAME_1);
            assertEquals(retrievedEndpoint.getEndpointName(), ENDPOINT_NAME_2);
            assertEquals(retrievedEndpoint.getLocation().getContainerName(), CONTAINER_NAME_2);

            // Check that only one endpoint (remote endpoint) is registered as implementing the service
            final Collection<PetalsServiceEndpoint> svcEpForSvc = localRegistryOverlayImpl
                    .getEndpointsForService(SERVICE_NAME_1, null);
            assertNotNull("Null list returned", svcEpForSvc);
            assertEquals("No endpoint retrieved for service", svcEpForSvc.size(), 1);
            assertEquals(svcEpForSvc.iterator().next().getEndpointName(), ENDPOINT_NAME_2);

            // Check that only one endpoint (remote endpoint) is registered as implementing the interface
            final Collection<PetalsServiceEndpoint> svcEpForItf = localRegistryOverlayImpl
                    .getEndpointsForInterface(INTERFACE_NAME_1, null);
            assertNotNull("Null list returned", svcEpForItf);
            assertEquals("No endpoint retrieved for interface", svcEpForItf.size(), 1);
            assertEquals(svcEpForItf.iterator().next().getEndpointName(), ENDPOINT_NAME_2);

            // ///////////////////////////////
            // Reactivate endpoints removed
            // ///////////////////////////////
            localRegistryOverlayImpl.joinTopology(removedEndpoints);

            assertEndpointsRegistered(localRegistryOverlayImpl);

        } finally {
            remoteRegistryOverlayImpl.stop();
            localRegistryOverlayImpl.stop();
        }
    }

    private static final void assertEndpointsRegistered(final RegistryOverlayImpl localRegistryOverlayImpl)
            throws SharedAreaException, WSDL4ComplexWsdlException, URISyntaxException {

        // Check that the remote endpoint is visible locally
        final PetalsServiceEndpoint retrievedRemoteEndpoint = localRegistryOverlayImpl.getEndpoint(SERVICE_NAME_1,
                ENDPOINT_NAME_2);
        assertNotNull(retrievedRemoteEndpoint);
        assertEquals(retrievedRemoteEndpoint.getServiceName(), SERVICE_NAME_1);
        assertEquals(retrievedRemoteEndpoint.getEndpointName(), ENDPOINT_NAME_2);
        assertEquals(retrievedRemoteEndpoint.getLocation().getContainerName(), CONTAINER_NAME_2);
        assertEndpointDescriptionParsable(localRegistryOverlayImpl, retrievedRemoteEndpoint);

        // Check the registration of local endpoint
        final PetalsServiceEndpoint retrievedLocalEndpoint = localRegistryOverlayImpl.getEndpoint(SERVICE_NAME_1,
                ENDPOINT_NAME_1);
        assertNotNull(retrievedLocalEndpoint);
        assertEquals(retrievedLocalEndpoint.getServiceName(), SERVICE_NAME_1);
        assertEquals(retrievedLocalEndpoint.getEndpointName(), ENDPOINT_NAME_1);
        assertEquals(retrievedLocalEndpoint.getLocation().getContainerName(), CONTAINER_NAME_1);
        assertEndpointDescriptionParsable(localRegistryOverlayImpl, retrievedLocalEndpoint);

        // Check that two endpoints are registered as implementing the service
        final Collection<PetalsServiceEndpoint> registeredSvcEpForSvc = localRegistryOverlayImpl
                .getEndpointsForService(SERVICE_NAME_1, null);
        assertNotNull("Null list returned", registeredSvcEpForSvc);
        assertEquals("No endpoint retrieved for service", registeredSvcEpForSvc.size(), 2);

        // Check that two endpoints are registered as implementing the interface
        final Collection<PetalsServiceEndpoint> registeredSvcEpForItf = localRegistryOverlayImpl
                .getEndpointsForInterface(INTERFACE_NAME_1, null);
        assertNotNull("Null list returned", registeredSvcEpForItf);
        assertEquals("No endpoint retrieved for interface", registeredSvcEpForItf.size(), 2);
    }

}
